/* exs_nc.c */

/* netcat for EXS.  See README in this directory for a summary of the intent and
 * architecture of this example program. */

/*
 * Copyright 2014 by
 *	University of New Hampshire Interoperability Laboratory
 *	121 Technology Drive, Suite 2
 *	Durham, NH 03824-4716
 *	phone: 1-603-862-0090
 *	fax: 1-603-862-4181
 *	e-mail: info@iol.unh.edu
 * All rights reserved.
 *
 * This file is part of the UNH EXS implementation of the
 * Consortium Specification: Extended Sockets API (ES-API) Issue 1.0,
 * Document Number C050 published by The Open Group, January 2005,
 * and copyright 2005 by The Open Group.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING.GPL in the main directory of this source tree, or the
 * BSD license, available from the file COPYING.BSD in the main directory
 * of this source tree.
 *
 * This material is based upon work supported by the National Science
 * Foundation under Grant No. OCI-1127228 and Grant No. DGE-0913620.
 *
 * Any opinions, findings, and conclusions or recommendations
 * expressed in this material are those of the authors and do not
 * necessarily reflect the views of the National Science Foundation.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#define _POSIX_C_SOURCE 200112L
#define _XOPEN_SOURCE 600

#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <pthread.h>
#include <unistd.h>

#include <exs.h>
#include "exs_extra.h"

const char *default_port = "31337";

static void
usage(void)
{
	fprintf(stderr, "Usage: exs_nc [options] [hostname] [port]\n");
	fprintf(stderr, "-4\tListen or connect using IPv4 only\n");
	fprintf(stderr, "-6\tListen or connect using IPv6 only\n");
	fprintf(stderr, "-d\tDo not attempt to read from stdin\n");
	fprintf(stderr, "-k\tKeep listening for more connections\n");
	fprintf(stderr, "-l\tBind and listen for incoming connections\n");
	fprintf(stderr, "-h\tPrint this help message\n");
	fprintf(stderr, "-q\tUse message-oriented transport\n");
}

static int (*listener_handlers[])(exs_event_t*);
static int (*connect_handlers[])(exs_event_t*);

/* Connection state.  Connections are initially in the unbound state.  Data
 * are sent on the connection only in the running state.  When the connection
 * needs to close, the state is changed to closing, but the actual closing of
 * the socket happens asynchronously. */
enum conn_state {
	conn_state_unbound,
	conn_state_running,
	conn_state_closing
};

/* We use a generic context object with three different subtypes, one for the
 * listener, one for connections (used for both client and server), and one for
 * the stdin thread.  The context object is passed into the "ahandle" field of
 * EXS calls so that it gets passed back in completion events and can be used to
 * retrieve/change the connection state. */
enum context_type_tag {
	context_listener,
	context_conn,
	context_stdin,
};

/* Connection context object.  Tracks connection state, parameters, and the
 * receive buffer for the connection. */
struct conn_context {
	int state;
	int fd;
	size_t nbufs;
	size_t bufsize;

	char *recv_bufspace;
	exs_mhandle_t recv_mhandle;
};

/* Listener context object.  Tracks the number of active and total listening
 * sockets, as well as the remaining number of connections that can be made,
 * after which the listeners will shut down. */
struct listener_context {
	size_t nfds;
	size_t nfds_alive;
	int *fds;
	int max_connections;
	int reserved_;
};

/* Stdin thread context object.  This is used for synchronization and also
 * contains the send buffers, which are shared by all connections (we follow the
 * convention of nmap's netcat implementation in which data from stdin is sent
 * on all open connections, i.e., poor man's multicast).  Each connection must
 * register with the stdin thread in order for data to be sent on that
 * connection. */
struct stdin_context {
	pthread_mutex_t send_mutex;
	pthread_cond_t send_avail;
	char *send_bufspace;
	exs_mhandle_t send_mhandle;
	char *send_start;
	int *send_pending_count;
	struct context **conns;
	size_t nfds_max;
	size_t nfds;
	size_t nbufs;
	size_t bufsize;
};

/* Wrapper for all context objects.  Does reference counting, and implements a
 * tree-like structure, so that connection objects have a listener as their
 * parent, etc. */
struct context {
	pthread_mutex_t ref_mutex;
	int refcount;
	int type;
	exs_qhandle_t qhandle;
	struct context *parent;
	union {
		struct listener_context lis;
		struct conn_context conn;
		struct stdin_context in;
	};
};

enum flags {
	/* Do not create a stdin thread */
	flag_no_stdin = 1 << 0,

	/* Listen for connections indefinitely */
	flag_keep_listening = 1 << 1,

	/* Shorthand for all defined flags */
	flag_all = flag_no_stdin | flag_keep_listening,
};

/* Structure to keep track of command-line options. */
struct options {
	int flags;
	int family;
	int socktype;
};


/* Increment the reference count of a context object.  No-op if the context
 * object is NULL. */
static struct context *
context_ref(struct context *ctx)
{
	if (ctx) {
		pthread_mutex_lock(&ctx->ref_mutex);
		++ctx->refcount;
		pthread_mutex_unlock(&ctx->ref_mutex);
	}
	return ctx;
}	/* context_ref */


#define NBUFS_DEFAULT 32
#define BUFSIZE_DEFAULT 4096

#ifdef PAGE_SIZE
#define MY_PAGE_SIZE PAGE_SIZE
#else
#define MY_PAGE_SIZE 4096
#endif

/* Creates a context for a listener. */
static struct context *
listener_context_new(int *fds, int nfds, struct context *parent,
							exs_qhandle_t qhandle)
{
	struct context *ctx;
	int ret;

	if (!(ctx = calloc(1, sizeof(*ctx)))) {
		goto out;
	}

	ret = pthread_mutex_init(&ctx->ref_mutex, NULL);
	if (ret) {
		errno = ret;
		goto err1;
	}

	ctx->refcount = 1;
	ctx->qhandle = qhandle;
	ctx->type = context_listener;
	ctx->parent = context_ref(parent);
	ctx->lis.fds = fds;
	ctx->lis.nfds = ctx->lis.nfds_alive = nfds;
	ctx->lis.max_connections = 1;
	goto out;

err1:
	free(ctx);
	ctx = NULL;
out:
	return ctx;
}	/* listener_context_new */

/* Creates a context for a new connection.  This creates and registers a large
 * memory area to receive messages on the connection. */
static struct context *
conn_context_new(int fd, struct context *parent, exs_qhandle_t qhandle)
{
	struct context *ctx;
	size_t totalsize;
	int ret;

	if (!(ctx = calloc(1, sizeof(*ctx)))) {
		goto out;
	}
	ctx->type = context_conn;
	ctx->qhandle = qhandle;
	ctx->conn.nbufs = NBUFS_DEFAULT;
	ctx->conn.bufsize = BUFSIZE_DEFAULT;

	totalsize = ctx->conn.nbufs * ctx->conn.bufsize;
	ret = posix_memalign((void **)&ctx->conn.recv_bufspace, MY_PAGE_SIZE, totalsize);
	if (ret) {
		errno = ret;
		goto err1;
	}

	ctx->conn.recv_mhandle = exs_mregister(ctx->conn.recv_bufspace,
								totalsize, 0);
	if (ctx->conn.recv_mhandle == EXS_MHANDLE_INVALID) {
		goto err2;
	}

	ret = pthread_mutex_init(&ctx->ref_mutex, NULL);
	if (ret) {
		errno = ret;
		goto err3;
	}

	if ((ctx->parent = context_ref(parent))) {
		if (parent->type == context_listener) {
			parent->lis.nfds_alive++;
		}
	}
	ctx->conn.fd = fd;
	ctx->refcount = 1;
	goto out;

err3:
	exs_mderegister(ctx->conn.recv_mhandle, 0);
err2:
	free(ctx->conn.recv_bufspace);
err1:
	free(ctx);
	ctx = NULL;
out:
	return ctx;
}	/* conn_context_new */


/* Creates a new stdin context object with the given number of send buffers of
 * the given size.  This function creates and allocates the send buffers. */
static struct context *
stdin_context_new(size_t nbufs, size_t bufsize)
{
	struct context *ctx;
	int ret;

	if (!(ctx = calloc(1, sizeof(*ctx)))) {
		goto out;
	}

	ret = posix_memalign((void **)&ctx->in.send_bufspace, MY_PAGE_SIZE,
			nbufs * bufsize);
	if (ret) {
		errno = ret;
		goto err0;
	}

	if (!(ctx->in.send_pending_count = calloc(nbufs,
					sizeof(*ctx->in.send_pending_count)))) {
		goto err1;
	}

	ctx->in.nfds_max = 100;
	if (!(ctx->in.conns = calloc(ctx->in.nfds_max, sizeof(*ctx->in.conns)))) {
		goto err2;
	}

	ctx->in.send_mhandle = exs_mregister(ctx->in.send_bufspace,
					nbufs * bufsize, EXS_MRF_RECV_DISABLE);
	if (ctx->in.send_mhandle == EXS_MHANDLE_INVALID) {
		errno = ret;
		goto err3;
	}

	ret = pthread_mutex_init(&ctx->in.send_mutex, NULL);
	if (ret) {
		errno = ret;
		goto err4;
	}

	ret = pthread_mutex_init(&ctx->ref_mutex, NULL);
	if (ret) {
		errno = ret;
		goto err5;
	}

	ret = pthread_cond_init(&ctx->in.send_avail, NULL);
	if (ret) {
		errno = ret;
		goto err6;
	}

	ctx->type = context_stdin;
	ctx->qhandle = EXS_QHANDLE_INVALID;
	ctx->in.send_start = ctx->in.send_bufspace;
	ctx->in.nbufs = nbufs;
	ctx->in.bufsize = bufsize;
	goto out;

err6:
	pthread_mutex_destroy(&ctx->ref_mutex);
err5:
	pthread_mutex_destroy(&ctx->in.send_mutex);
err4:
	exs_mderegister(ctx->in.send_mhandle, 0);
err3:
	free(ctx->in.conns);
err2:
	free(ctx->in.send_pending_count);
err1:
	free(ctx->in.send_bufspace);
err0:
	free(ctx);
	ctx = NULL;
out:
	return ctx;
}	/* stdin_context_new */


/* Posts all of the receive buffers on the given connection.  Intended to be
 * called at connection startup time. */
static int
conn_start(struct context *ctx)
{
	char *buf, *end;
	int ret;

	assert(ctx->type == context_conn);
	ctx->conn.state = conn_state_running;
	for (buf = ctx->conn.recv_bufspace, end = buf + ctx->conn.nbufs * ctx->conn.bufsize;
					buf != end; buf += ctx->conn.bufsize) {
		ret = exs_recv(ctx->conn.fd, buf, ctx->conn.bufsize, 0,
				ctx->qhandle, (exs_ahandle_t)ctx,
				ctx->conn.recv_mhandle);
		if (ret) {
			if (errno == EBUSY) {
				break;
			}
			return -1;
		}
	}
	return 0;
}	/* conn_start */


/* Asynchronously closes the given connection.  The flags are passed as-is to
 * exs_close(), which is used to pass EXS_DONTLINGER if we need to abruptly
 * terminate the connection. */
static int
conn_start_close(struct context *ctx, int flags)
{
	int ret;

	assert(ctx->type == context_conn);
	assert(ctx->conn.fd >= 0);
	if (ctx->conn.state == conn_state_closing) {
		/* Close already issued */
		return 0;
	}
	ctx->conn.state = conn_state_closing;
#ifdef DEBUG
	fprintf(stderr, "Start close on fd %d\n", ctx->conn.fd);
#endif
	if ((ret = exs_close(ctx->conn.fd, flags, ctx->qhandle, ctx)) < 0) {
		fprintf(stderr, "Could not close fd %d\n", ctx->conn.fd);
		return -1;
	} else {
		return 1;
	}
}	/* conn_start_close */


/* Frees the listener context.  This operation is only meant to be called from
 * the context_unref() function. */
static void
listener_context_free(struct context *ctx)
{
	if (!ctx) {
		return;
	}

	assert(ctx->refcount == 0);
	assert(ctx->type == context_conn);
	pthread_mutex_destroy(&ctx->ref_mutex);
	free(ctx->lis.fds);
	exs_qdelete(ctx->qhandle);
	free(ctx);
}	/* listener_context_free */


/* Frees a connection context.  This operation is only meant to be called from
 * the context_unref() function. */
static void
conn_context_free(struct context *ctx)
{
	if (!ctx) {
		return;
	}

	assert(ctx->refcount == 0);
	assert(ctx->type == context_conn);
	assert(ctx->conn.state == conn_state_closing);
	pthread_mutex_destroy(&ctx->ref_mutex);
	if (ctx->conn.recv_mhandle != EXS_MHANDLE_INVALID) {
		exs_mderegister(ctx->conn.recv_mhandle, 0);
	}
	if (ctx->conn.recv_bufspace) {
		free(ctx->conn.recv_bufspace);
	}
	if (!ctx->parent) {
		exs_qdelete(ctx->qhandle);
	}
	free(ctx);
}	/* conn_context_free */


/* Frees a stdin context.  This operation is only meant to be called from the
 * context_unref() function. */
static void
stdin_context_free(struct context *ctx)
{
	if (!ctx) {
		return;
	}

	assert(ctx->refcount == 0);
	assert(ctx->type == context_stdin);
	pthread_mutex_destroy(&ctx->ref_mutex);
	pthread_mutex_destroy(&ctx->in.send_mutex);
	if (ctx->in.send_mhandle != EXS_MHANDLE_INVALID) {
		exs_mderegister(ctx->in.send_mhandle, 0);
	}
	if (ctx->in.conns) {
		free(ctx->in.conns);
	}
	if (ctx->in.send_pending_count) {
		free(ctx->in.send_pending_count);
	}
	if (ctx->in.send_bufspace) {
		free(ctx->in.send_bufspace);
	}
	free(ctx);
}	/* stdin_context_free */


/* Decrements the reference count on the given context object and frees the
 * context object if the reference count becomes 0 as a result.  It is an error
 * if this function is passed NULL. */
static struct context *
context_unref(struct context *ctx)
{
	int new_refcount;
	assert(ctx != NULL);
	pthread_mutex_lock(&ctx->ref_mutex);
	new_refcount = --ctx->refcount;
	pthread_mutex_unlock(&ctx->ref_mutex);
	if (new_refcount == 0) {
		switch ((enum context_type_tag)ctx->type) {
		case context_listener:
			listener_context_free(ctx);
			break;
		case context_conn:
			conn_context_free(ctx);
			break;
		case context_stdin:
			stdin_context_free(ctx);
			break;
		}
	}
	return NULL;
}	/* context_unref */


/* Registers a connection with the stdin reader thread.  This will cause
 * standard input data to be sent on the given connection. */
static int
stdin_register(struct context *in_ctx, struct context *conn_ctx)
{
	int signal;

	if (!in_ctx) {
		/* Ignore if no stdin thread running. */
		return 0;
	}
	assert(in_ctx->type == context_stdin);
	assert(conn_ctx && conn_ctx->type == context_conn);

	if (in_ctx->in.nfds >= in_ctx->in.nfds_max) {
		return -ENOBUFS;
	}

	pthread_mutex_lock(&in_ctx->in.send_mutex);
	signal = !in_ctx->in.nfds;
	in_ctx->in.conns[in_ctx->in.nfds++] = context_ref(conn_ctx);
	pthread_mutex_unlock(&in_ctx->in.send_mutex);
	if (signal) {
		pthread_cond_signal(&in_ctx->in.send_avail);
	}
	return 0;
}	/* stdin_register */


/* Deregisters a connection with the stdin reader thread.  After this function
 * completes, no more data from stdin will be sent on the given connection. */
static int
stdin_deregister(struct context *conn_ctx)
{
	struct context *in_ctx;
	struct context **iter, **end;
	int ret = -EBADF;

	assert(conn_ctx && conn_ctx->type == context_conn);

	in_ctx = conn_ctx->parent;
	while (in_ctx && in_ctx->type != context_stdin) {
		in_ctx = in_ctx->parent;
	}
	if (!in_ctx) {
		/* Ignore request if no stdin thread running. */
		return 0;
	}
	assert(in_ctx->type == context_stdin);

	pthread_mutex_lock(&in_ctx->in.send_mutex);
	for (iter = in_ctx->in.conns, end = iter + in_ctx->in.nfds;
							iter != end; ++iter) {
		if (*iter == conn_ctx) {
			*iter = in_ctx->in.conns[--in_ctx->in.nfds];
			ret = 0;
			break;
		}
	}
	pthread_mutex_unlock(&in_ctx->in.send_mutex);

	return ret;
}	/* stdin_deregister */


/* Shuts down all connections for writing.  Called only from
 * stdin_reader_thread when EOF is reached or there is an error reading from
 * stdin. */
static void
stdin_close_all(struct context *in_ctx, int flags)
{
	struct context **iter, **end;

	assert(in_ctx && in_ctx->type == context_stdin);

	pthread_mutex_lock(&in_ctx->in.send_mutex);
	for (iter = in_ctx->in.conns, end = iter + in_ctx->in.nfds;
							iter != end; ++iter) {
		if (flags) {
			conn_start_close(*iter, flags);
		} else {
			if (exs_shutdown((*iter)->conn.fd, SHUT_WR, 0,
					(*iter)->qhandle, *iter) < 0) {
				conn_start_close(*iter, flags);
			}
		}
		*iter = context_unref(*iter);
	}
	pthread_mutex_unlock(&in_ctx->in.send_mutex);
}	/* stdin_close_all */


/* Reads a block of data from stdin and then sends the data on all
 * registered connections.  The next send block must be available---i.e., there
 * must not already be a pending operation on the block. */
static int
pull_from_stdin(struct context *ctx)
{
	struct context **iter;
	int index;
	int ret;

	assert(ctx->type == context_stdin);
	ret = read(STDIN_FILENO, ctx->in.send_start, ctx->in.bufsize);
	if (ret < 0 && errno != EAGAIN && errno != EWOULDBLOCK) {
		fprintf(stderr, "Could not read from stdin: %s\n",
				strerror(errno));
		return ret;
	} else if (ret <= 0) {
		return ret;
	}

	pthread_mutex_lock(&ctx->in.send_mutex);
	index = (ctx->in.send_start - ctx->in.send_bufspace) / ctx->in.bufsize;
	for (iter = ctx->in.conns; iter != ctx->in.conns + ctx->in.nfds; ++iter) {
		if (exs_send((*iter)->conn.fd, ctx->in.send_start, ret, 0,
					(*iter)->qhandle, *iter,
					ctx->in.send_mhandle) < 0) {
			fprintf(stderr, "Start send failed on socket %d: %s\n",
					(*iter)->conn.fd, strerror(errno));
			ret = -1;
			conn_start_close(ctx, EXS_DONTLINGER);
			/* Deregister this connection */
			*iter = ctx->in.conns[--ctx->in.nfds];
			/* Try sending on the conn that we swapped with now */
			--iter;
		}
	}
	if (iter != ctx->in.conns) {
		ctx->in.send_start += ctx->in.bufsize;
		if (ctx->in.send_start - ctx->in.send_bufspace
				>= (ptrdiff_t)(ctx->in.nbufs * ctx->in.bufsize)) {
			ctx->in.send_start = ctx->in.send_bufspace;
		}
		ctx->in.send_pending_count[index] += iter - ctx->in.conns;
	}
	pthread_mutex_unlock(&ctx->in.send_mutex);
	return ret;
}	/* pull_from_stdin */


/* A thread which reads data from standard input until EOF, and sends the data
 * on all connections.  The thread waits if there are no connections or if at
 * least one connection is busy sending data. */
static void *
stdin_reader_thread(void *in)
{
	struct context *ctx = context_ref((struct context *)in);
	unsigned int index;
	int ret;

	pthread_detach(pthread_self());

	do {
		pthread_mutex_lock(&ctx->in.send_mutex);
		index = (ctx->in.send_start - ctx->in.send_bufspace)
							/ ctx->in.bufsize;
		while (ctx->in.nfds == 0
				|| ctx->in.send_pending_count[index] > 0) {
			pthread_cond_wait(&ctx->in.send_avail,
							&ctx->in.send_mutex);
		}
		pthread_mutex_unlock(&ctx->in.send_mutex);
	} while ((ret = pull_from_stdin(ctx)) > 0);
	if (ret < 0) {
		perror("read from stdin");
	}
	stdin_close_all(ctx, (ret < 0) ? EXS_DONTLINGER : 0);
	ctx = context_unref(ctx);
#ifdef DEBUG
	fprintf(stderr, "Leaving reader thread\n");
#endif
	return NULL;
}	/* stdin_reader_thread */


/* Event handler which handles an unexpected EXS event type. */
static int
unexpected_handler(exs_event_t *evt)
{
	fprintf(stderr, "Got unexpected event: %s\n",
			exs_evt_type_to_string(evt->exs_evt_type));
	exit(EXIT_FAILURE);
}	/* unexpected_handler */


/* Event handler which handles an EXS connect event.  Starts receives on the
 * connection and registers the connection with the stdin reader thread. */
static int
connect_handler(exs_event_t *evt)
{
	struct context *ctx = evt->exs_evt_ahandle;
	int ret;

	assert(ctx->type == context_conn);
	if (evt->exs_evt_errno != 0) {
		fprintf(stderr, "Connect failed on socket %d: %s\n",
				evt->exs_evt_socket,
				strerror(evt->exs_evt_errno));
		return -1;
	}

	ret = conn_start(ctx);
	if (ret) {
		stdin_deregister(ctx);
		return conn_start_close(ctx, EXS_DONTLINGER);
	}

	return 1;
}	/* connect_handler */


/* Event handler which handles an EXS recv event.  Reposts the receive and
 * outputs the received data to standard output (we assume that this will always
 * succeed for now). */
static int
recv_handler(exs_event_t *evt)
{
	struct context *ctx = evt->exs_evt_ahandle;
	char *buf;
	size_t length;
	ssize_t ret;

	assert(ctx->type == context_conn);
	if (evt->exs_evt_errno != 0) {
		fprintf(stderr, "Recv failed on socket %d: %s\n",
				evt->exs_evt_socket,
				strerror(evt->exs_evt_errno));
		stdin_deregister(ctx);
		return conn_start_close(ctx, EXS_DONTLINGER);
	}
	buf = evt->exs_evt_union.exs_evt_xfer.exs_evt_buffer;
	length = evt->exs_evt_union.exs_evt_xfer.exs_evt_length;

	if (length == 0) {
		/* Got EOF */
		stdin_deregister(ctx);
		return conn_start_close(ctx, 0);
	}

	/* Output what we just received */
	if ((ret = write(STDOUT_FILENO, buf, length)) < 0) {
		if (errno == EAGAIN) {
			fprintf(stderr, "Saturated standard output buffer!\n");
			fprintf(stderr, "TODO: Do something reasonable here\n");
		} else {
			fprintf(stderr, "Could not write to stdout: %s\n",
							strerror(errno));
		}
		stdin_deregister(ctx);
		return conn_start_close(ctx, EXS_DONTLINGER);
	}

	/* Receive next buffer */
	if (exs_recv(evt->exs_evt_socket, buf, ctx->conn.bufsize, 0,
				ctx->qhandle, ctx,
				ctx->conn.recv_mhandle) < 0) {
		fprintf(stderr, "Repost recv failed on socket %d: %s\n",
				evt->exs_evt_socket, strerror(errno));
		stdin_deregister(ctx);
		return conn_start_close(ctx, EXS_DONTLINGER);
	}
	return 1;
}	/* recv_handler */


/* Handles an EXS send event.  Update the stdin reader thread state accordingly,
 * to allow the thread to reuse the send buffer. */
static int
send_handler(exs_event_t *evt)
{
	struct context *ctx = evt->exs_evt_ahandle;
	struct context *in_ctx = ctx->parent;
	char *buf = evt->exs_evt_union.exs_evt_xfer.exs_evt_buffer;
	unsigned int index;
	int dont_signal;
	while (in_ctx && in_ctx->type != context_stdin) {
		in_ctx = in_ctx->parent;
	}
	assert(in_ctx && in_ctx->type == context_stdin);
	assert(ctx->type == context_conn);
	pthread_mutex_lock(&in_ctx->in.send_mutex);
	index = (buf - in_ctx->in.send_bufspace) / in_ctx->in.bufsize;
	assert(index < in_ctx->in.nbufs);
	assert(in_ctx->in.send_pending_count[index] > 0);
	dont_signal = --in_ctx->in.send_pending_count[index];
	pthread_mutex_unlock(&in_ctx->in.send_mutex);
	if (!dont_signal) {
		pthread_cond_signal(&in_ctx->in.send_avail);
	}
	if (evt->exs_evt_errno != 0) {
		fprintf(stderr, "Send failed on socket %d: %s\n",
				evt->exs_evt_socket, strerror(evt->exs_evt_errno));
		stdin_deregister(ctx);
		return conn_start_close(ctx, EXS_DONTLINGER);
	}
	return 0;
}	/* send_handler */


/* Handles an EXS accept event.  Creates a context object for the new
 * connection, registers the new connection with the stdin reader thread and
 * posts receives on the connection. */
static int
accept_handler(exs_event_t *evt)
{
	struct context *ctx = evt->exs_evt_ahandle;
	struct context *newctx;
	exs_acceptaddr_t addr;
	unsigned int i;
	int newfd, ret;

	assert(ctx->type == context_listener);
	if (evt->exs_evt_errno != 0) {
		fprintf(stderr, "Error accepting connection on socket %d: %s\n",
				evt->exs_evt_socket,
				strerror(evt->exs_evt_errno));
		return 0;
	}
	newfd = evt->exs_evt_union.exs_evt_accept.exs_evt_new_socket;
#ifdef DEBUG
	fprintf(stderr, "Accepted new connection on socket %d, from socket %d\n",
			newfd,
			evt->exs_evt_socket);
#endif
	if (!(newctx = conn_context_new(newfd, ctx, ctx->qhandle))) {
		fprintf(stderr, "Error: could not create context for socket %d: %s\n",
				evt->exs_evt_socket, strerror(errno));
		if (exs_close(newfd, 0, ctx->qhandle, ctx) < 0) {
			fprintf(stderr, "Warning: error closing fd %d: %s\n",
						newfd, strerror(errno));
			return 0;
		}
		return 1;
	}
	if (stdin_register(ctx->parent, newctx) < 0) {
		fprintf(stderr, "Could not register socket %d for stdin reading: %s\n",
				evt->exs_evt_socket, strerror(errno));
		stdin_deregister(ctx);
		return conn_start_close(newctx, EXS_DONTLINGER);
	}
	if ((conn_start(newctx)) < 0) {
		fprintf(stderr, "Could not start recv requests on socket %d: %s\n",
				evt->exs_evt_socket, strerror(errno));
		stdin_deregister(ctx);
		return conn_start_close(newctx, EXS_DONTLINGER);
	}

	ret = newctx->conn.nbufs;
	if (ctx->lis.max_connections < 0 || --ctx->lis.max_connections) {
		/* Repost this exs_accept operation */
		addr.exs_addr = NULL;
		addr.exs_addrlen = 0;
		addr.exs_ahandle = evt->exs_evt_ahandle;
		fprintf(stderr, "new ahandle %p\n", addr.exs_ahandle);
		if (evt->exs_evt_errno == 0 && exs_accept(evt->exs_evt_socket,
						&addr, 1, 0, ctx->qhandle) < 0) {
			fprintf(stderr, "Cannot accept connections on fd %d: %s\n",
						evt->exs_evt_socket,
						strerror(errno));
		}
		ret++;
	} else {
		for (i = 0; i < ctx->lis.nfds; ++i) {
			if (exs_close(ctx->lis.fds[i], 0, ctx->qhandle, ctx) < 0) {
				fprintf(stderr, "Warning: error closing fd %d: %s\n",
						ctx->lis.fds[i], strerror(errno));
			} else {
				ret++;
			}
		}
	}
	return ret;
}	/* accept_handler */


/* Handles an EXS shutdown event.  This currently does not do much. */
static int
shutdown_handler(exs_event_t *evt)
{
	int ret = 0;
	if (evt->exs_evt_errno != 0) {
		fprintf(stderr, "Error shutting down socket %d for writing: %s\n",
				evt->exs_evt_socket,
				strerror(evt->exs_evt_errno));
	}
	fflush(stdout);
	return ret;
}	/* shutdown_handler */


/* Handles an EXS close event.  Dereferences the associated context object,
 * and walks up the tree structure to see if this is the last socket.  If this
 * is the last socket, indicate that the event loop should terminate. */
static int
close_handler(exs_event_t *evt)
{
	struct context *ctx = evt->exs_evt_ahandle;
	struct context *parent;
	int ret = 0;
	if (evt->exs_evt_errno != 0) {
		fprintf(stderr, "Error closing socket %d: %s\n",
				evt->exs_evt_socket,
				strerror(evt->exs_evt_errno));
	}
	fflush(stdout);
	while (ctx) {
		if (ctx->type == context_listener) {
			if ((--ctx->lis.nfds_alive)) {
				/* still more fds using this context;
				 * keep it aroud a bit longer */
#ifdef DEBUG
				fprintf(stderr, "nfds now %zu\n", ctx->lis.nfds_alive);
#endif
				break;
			} else {
				/* get out of event loop */
#ifdef DEBUG
				fprintf(stderr, "nfds now %zu; getting out\n", ctx->lis.nfds_alive);
#endif
				ret = -1;
			}
		} else if (!ctx->parent) {
			ret = -1;
		}
		parent = ctx->parent;
		context_unref(ctx);
		ctx = parent;
	}
	return ret;
}	/* close_handler */


/* Client event handler array. */
static int (*connect_handlers[])(exs_event_t*) = {
	[EXS_EVT_POLL] = &unexpected_handler,
	[EXS_EVT_CONNECT] = &connect_handler,
	[EXS_EVT_ACCEPT] = &unexpected_handler,
	[EXS_EVT_SEND] = &send_handler,
	[EXS_EVT_RECV] = &recv_handler,
	[EXS_EVT_SENDMSG] = &unexpected_handler,
	[EXS_EVT_RECVMSG] = &unexpected_handler,
	[EXS_EVT_SENDFILE] = &unexpected_handler,
	[EXS_EVT_SENDTIMEOUT] = &unexpected_handler,
	[EXS_EVT_RECVTIMEOUT] = &unexpected_handler,
	[EXS_EVT_CLOSE] = &close_handler,
	[EXS_EVT_SHUTDOWN] = &shutdown_handler,
};


/* Server event handler array. */
static int (*listener_handlers[])(exs_event_t*) = {
	[EXS_EVT_POLL] = &unexpected_handler,
	[EXS_EVT_CONNECT] = &unexpected_handler,
	[EXS_EVT_ACCEPT] = &accept_handler,
	[EXS_EVT_SEND] = &send_handler,
	[EXS_EVT_RECV] = &recv_handler,
	[EXS_EVT_SENDMSG] = &unexpected_handler,
	[EXS_EVT_RECVMSG] = &unexpected_handler,
	[EXS_EVT_SENDFILE] = &unexpected_handler,
	[EXS_EVT_SENDTIMEOUT] = &unexpected_handler,
	[EXS_EVT_RECVTIMEOUT] = &unexpected_handler,
	[EXS_EVT_CLOSE] = &close_handler,
	[EXS_EVT_SHUTDOWN] = &shutdown_handler,
};


/* The main program for the client.  Sets up the stdin reader thread and
 * starts the connection to the endpoint requested by the user. */
static int
run_connect_exs(const char *hostname, const char *port, struct options *options)
{
	struct context *ctx;
	struct context *inctx;
	exs_qhandle_t master_queue;
	exs_event_t event;
	pthread_t child;
	int ret;
	int nevents;

	errno = 0;
	master_queue = exs_qcreate(0);
	if (master_queue == EXS_QHANDLE_INVALID) {
		perror("Could not create event queue");
		exit(EXIT_FAILURE);
	}

	if (options->flags & flag_no_stdin) {
		inctx = NULL;
	} else if (!(inctx = stdin_context_new(32, 4096))) {
			perror("Could not create stdin context");
			exit(EXIT_FAILURE);
	}

	if (!(ctx = conn_context_new(-1, inctx, master_queue))) {
		perror("Could not create connection context");
		exit(EXIT_FAILURE);
	}

	if ((ctx->conn.fd = exs_connect_new(hostname, port,
					options->family, options->socktype, 0,
					NULL, master_queue, ctx)) < 0) {
		fprintf(stderr, "Could not connect socket to %s:%s\n",
								hostname, port);
		exit(EXIT_FAILURE);
	}

	ret = exs_qdequeue(master_queue, &event, 1, NULL);
	if (ret < 0) {
		fprintf(stderr, "Error waiting for connect to complete: %s\n",
							strerror(errno));
		exit(EXIT_FAILURE);
	} else {
		assert(ret == 1);
		if (event.exs_evt_type != EXS_EVT_CONNECT) {
			unexpected_handler(&event);
			exit(EXIT_FAILURE);
		} else {
			if (connect_handler(&event) < 0) {
				exit(EXIT_FAILURE);
			}
		}
	}

	if (!(options->flags & flag_no_stdin)) {
		/* Register with stdin thread */
		stdin_register(inctx, ctx);

		/* Launch stdin reader thread */
		ret = pthread_create(&child, NULL, &stdin_reader_thread, inctx);
		if (ret) {
			fprintf(stderr, "Error creating stdin thread: %s\n",
							strerror(errno));
		}
	}

	nevents = 32;
	errno = 0;
	ret = dispatch_exs_events(master_queue, nevents,
				connect_handlers, NULL);
	if (ret < 0) {
		fprintf(stderr, "Error waiting for events: %s\n",
					strerror(ret));
		ret = EXIT_FAILURE;
	} else {
		ret = EXIT_SUCCESS;
	}

	return ret;
}	/* run_connect_exs */


/* The main program for the server.  Sets up the stdin reader thread and
 * listens on all available interfaces for incoming connections. */
int
run_listen_exs(const char *hostname, const char *port, struct options *options)
{
	struct context *ctx;
	struct context *in_ctx;
	exs_acceptaddr_t addr;
	exs_qhandle_t master_queue;
	pthread_t child;
	int i;
	int ret;
	int *fds;
	int nfds;
	int nevents;

	errno = 0;
	nfds = 0;
	fds = NULL;
	if (our_bind_all(hostname, port, options->family, options->socktype,
				&fds, &nfds) < 0) {
		fprintf(stderr, "Could not open listening socket: %s\n",
							strerror(errno));
		ret = EXIT_FAILURE;
		goto exit;
	} else if (nfds == 0) {
		fprintf(stderr, "No failure but did not bind to any interfaces\n");
		ret = EXIT_FAILURE;
		goto exit;
	}

	master_queue = exs_qcreate(0);
	if (master_queue == EXS_QHANDLE_INVALID) {
		perror("Could not create listener event queue");
		ret = EXIT_FAILURE;
		goto exit;
	}

	if ((options->flags & flag_no_stdin)) {
		in_ctx = NULL;
	} else {
		if (!(in_ctx = stdin_context_new(32, 4096))) {
			perror("Could not create stdin context");
			exit(EXIT_FAILURE);
		}
		ret = pthread_create(&child, NULL, &stdin_reader_thread, in_ctx);
		if (ret) {
			fprintf(stderr, "Error creating stdin thread: %s\n",
							strerror(errno));
			exit(EXIT_FAILURE);
		}
	}

	ctx = listener_context_new(fds, nfds, in_ctx, master_queue);
	if (!ctx) {
		perror("Could not create listener context");
		exit(EXIT_FAILURE);
	}
	if (options->flags & flag_keep_listening) {
		ctx->lis.max_connections = -1;
	}
	nevents = 0;
	for (i = 0; i < nfds; ++i) {
		addr.exs_addr = NULL;
		addr.exs_addrlen = 0;
		addr.exs_ahandle = (void *)ctx;
		if (exs_accept(fds[i], &addr, 1, 0, master_queue) < 0) {
			fprintf(stderr, "Cannot accept connections on fd %d: %s\n",
						fds[i], strerror(errno));
		}
		++nevents;
	}

	ret = dispatch_exs_events(master_queue, nevents, listener_handlers, NULL);
	ret = ret ? EXIT_FAILURE : EXIT_SUCCESS;

exit:
	return ret;
}	/* run_listen_exs */


int
main(int argc, char *argv[])
{
	const char *hostname = NULL;
	const char *port = default_port;
	char **arg;
	struct options options;
	enum mode { CONNECT_EXS, LISTEN_EXS } mode = CONNECT_EXS;
	int ch;

	port = "31337";
	memset(&options, 0, sizeof(options));
	options.family = AF_UNSPEC;
	options.socktype = SOCK_STREAM;
	while ((ch = getopt(argc, argv, "46dklhqQ")) >= 0) {
		switch (ch) {
		case '4':
			if (options.family != AF_UNSPEC) {
				fprintf(stderr, "Cannot use both -4 and -6\n");
				usage();
				exit(EXIT_FAILURE);
			}
			options.family = AF_INET;
			break;
		case '6':
			if (options.family != AF_UNSPEC) {
				fprintf(stderr, "Cannot use both -4 and -6\n");
				usage();
				exit(EXIT_FAILURE);
			}
			options.family = AF_INET6;
			break;
		case 'd':
			options.flags |= flag_no_stdin;
			break;
		case 'k':
			options.flags |= flag_keep_listening;
			break;
		case 'l':
			mode = LISTEN_EXS;
			break;
		case 'h':
			usage();
			exit(EXIT_SUCCESS);
		case 'q':
		case 'Q':
			options.socktype = SOCK_SEQPACKET;
			break;
		default:
			usage();
			exit(EXIT_FAILURE);
		}
	}

	if (((options.flags & flag_keep_listening)) && mode != LISTEN_EXS) {
		fprintf(stderr, "%s: Cannot use -k in connect mode\n", argv[0]);
		usage();
		exit(EXIT_FAILURE);
	}

	arg = argv + optind;
	if (*arg) {
		hostname = *arg;
		++arg;
	} else if (mode != LISTEN_EXS) {
		fprintf(stderr, "%s: No host specified.\n", argv[0]);
		usage();
		exit(EXIT_FAILURE);
	}

	if (*arg) {
		port = *arg;
	}

	if (exs_init(EXS_VERSION1) < 0) {
		perror("Could not initialize EXS");
		exit(EXIT_FAILURE);
	}

	switch (mode) {
	case CONNECT_EXS:
		return run_connect_exs(hostname, port, &options);
	case LISTEN_EXS:
		return run_listen_exs(hostname, port, &options);
	}

	return EXIT_FAILURE;
}	/* main */
